% $Id: square-1.pl,v 1.1 1996/07/11 13:21:15 uwe Exp $
/*
Problem:
given: a square of side length n
       and all squares of side length n-1 to 1
       (one square per size)
       arrange these smaller squares
       on the n-square (without overlapping)
       such that the covered area is maximal
       
       for smal values of n (e.g < 20) the best solution
       is with the two squares of size n-1 and 1
       
       for larger values of n this is not an optimal solution
       
Strategie in this Program:
       constraint satisfaction
       initialy all moves for all squares are computed,
       then moves are (arbitrarily) selected and the remaning
       moves of all squares are computed,
       this is repeated until no more moves are possible
       
data structures:
       the set of squares that stil could be used is
       stored in a list.
       Each entry contains the square size and
       the set of possible moves.
       This set is partitioned. The partitions form a list
       of rectangles.
       Each rectangle is represented by two points,
       the lower left corner and the upper rigth corner
       (inclusive)

*/

trace_output(0).

initstate(Square_size,State) :-
	Square_size1 is Square_size - 1,
	initstate(Square_size,Square_size1,State).

initstate(_,0,[]).
initstate(Square_size,I,[Area | State1]) :-
	I > 0,
	Area_size is Square_size - I + 1,
	Area = square(I,[r(p(0,0),p(Area_size,Area_size))]),
	I1 is I - 1,
	initstate(Square_size,I1,State1).
   
   
select_move([Square | _],Max_size,move(Square_size,Position)) :-
	Square = square(Square_size,Possible_positions),
	Square_size < Max_size,
	select_corner(Possible_positions,Position).
   
select_move([_ | State],Max_size,Move) :-
	select_move(State,Max_size,Move).
   
select_corner([r(Position, _) | _],Position).
select_corner([_ | Possible_positions],Position) :-
	select_corner(Possible_positions,Position).
   
satisfy_all_constraints(_,[],[]).
satisfy_all_constraints(Move,[Square | State1],Newstate) :-
	satisfy_all_constraints(Move,State1,Newstate1),
	satisfy_one_constraint(Move,Square,Newstate1,Newstate).
   
satisfy_one_constraint(move(Square_size1,_),
                       square(Square_size,_),State,State) :-
	Square_size >= Square_size1.

satisfy_one_constraint(Move,
                       square(Square_size,Possible_positions),
                       State,
                       Newstate) :-
	Move = move(Square_size1,_), Square_size < Square_size1,
	restrict_possible_positions(Move,Square_size,Possible_positions,New_possible_positions),
	add_square(Square_size,New_possible_positions,State,Newstate).
   
add_square(_,[],State,State) :- !.
add_square(Square_size,Possible_positions,State,
           [square(Square_size,Possible_positions) | State]).                    
          
restrict_possible_positions(_,_,[],[]).
restrict_possible_positions(Move,Square_size,[Rectangle | Positions],New_positions) :-
	restrict_positions(Move,Square_size,Rectangle,New_pos1),
	restrict_possible_positions(Move,Square_size,Positions,New_pos2),
	append_positions(New_pos1,New_pos2,New_positions).
   
append_positions([],Pos2,Pos2).
append_positions([H|T],Pos2,[H|T1]) :- append_positions(T,Pos2,T1).

trace_restrict_positions(Move,Square_size,Rectangle) :-
	trace_output(2),
	write('restrict_positions   : '),write( Move),nl,
	write('for square with size : '),write( Square_size),nl,
	write('and positions        : '),write( Rectangle),nl,
	!.
trace_restrict_positions(_,_,_).

restrict_positions(Move,Square_size,Rectangle,Remaining_moves) :-
	trace_restrict_positions(Move,Square_size,Rectangle),
	Move = move(Moved_square_size,p(X,Y)),
	Square_size1 is Square_size -1,
	X0 is X - Square_size1,
	Y0 is Y - Square_size1,
	X1 is X + Moved_square_size,
	Y1 is Y + Moved_square_size,
	Impossible_moves = r(p(X0,Y0),p(X1,Y1)),
	sub_rectangle(Rectangle,Impossible_moves,Remaining_moves).
   

maxi(X,Y,X) :- X >= Y.
maxi(X,Y,Y) :- X <  Y.
maxp(p(X0,Y0),p(X1,Y1),p(X2,Y2)) :-
	maxi(X0,X1,X2),
	maxi(Y0,Y1,Y2).

mini(X,Y,X) :- X =< Y.
mini(X,Y,Y) :- X >  Y.
minp(p(X0,Y0),p(X1,Y1),p(X2,Y2)) :-
	mini(X0,X1,X2),
	mini(Y0,Y1,Y2).

intersect_rectangle(r(P0,Q0),r(P1,Q1),r(P2,Q2)) :-
	maxp(P0,P1,P2),
	minp(Q0,Q1,Q2).

empty_rectangle(r(p(X0,_),p(X1,_))) :- X0 > X1, !.
empty_rectangle(r(p(_,Y0),p(_,Y1))) :- Y0 > Y1.

trace_sub_rectangle(R1,R2,Rl) :-
	trace_output(2),
	write('rectangle  : '),write(R1),write(' - '),write(R2),nl,
	write('split into : '),write(Rl),nl,
	!.
trace_sub_rectangle(_,_,_).

sub_rectangle(R1,R2,Rl) :-
	intersect_rectangle(R1,R2,R3),
	split_rectangle(R1,R3,Rl),
	trace_sub_rectangle(R1,R2,Rl).
   
split_rectangle(R0,R0,[])   :- !.
split_rectangle(R0,R1,[R0]) :- empty_rectangle(R1), !.

split_rectangle(R0,R1,Rl) :-
	R0 = r(P0,P1), R1 = r(P2,_),
	P0 = p(X0,Y0), P2 = p(X2,_),
	X0 < X2,!,
	P1 = p(_,Y1),
	Left_side  = r(P0,p(X2,Y1)),
	Right_side = r(p(X2,Y0),P1),
	split_rectangle(Right_side,R1,Rl1),
	Rl = [Left_side | Rl1].

split_rectangle(R0,R1,Rl) :-
	R0 = r(P0,P1), R1 = r(P2,_),
	P0 = p(X0,Y0), P2 = p(X2,Y2),
	Y0 < Y2, !, X0 = X2,
	P1 = p(X1,_),
	Lower_side = r(P0,p(X1,Y2)),
	Upper_side = r(P2,P1),
	split_rectangle(Upper_side,R1,Rl1),
	Rl = [Lower_side | Rl1].

split_rectangle(R0,R1,Rl) :-
	R0 = r(P0,P1), R1 = r(P0,P3),
	P1 = p(X1,Y1), P3 = p(X3,_),
	X1 > X3, !,
	P0 = p(_,Y0),
	Left_side  = r(P0,p(X3,Y1)),
	Right_side = r(p(X3,Y0),P1),
	split_rectangle(Left_side,R1,Rl1),
	Rl = [Right_side | Rl1].

split_rectangle(R0,R1,Rl) :-
	R0 = r(P0,P1), R1 = r(P0,P3),
	P1 = p(X1,Y1), P3 = p(X1,Y3),
	Y1 > Y3, !,
	P0 = p(X0,_),
	Upper_side = r(p(X0,Y3),P1),
	Rl = [Upper_side].

out_movelist([],0).
out_movelist([Move | Moves],Move_no) :-
	out_movelist(Moves,Move_no1),
	Move_no is Move_no1 + 1,
	out_move(Move,Move_no).

out_move(move(Size,p(X,Y)),Move_no) :-
	write(Move_no),
	write('. move : square of size = '),
	write(Size),
	write(' at ('),write(X),write(','),write(Y),write(')'),
	nl.

out_solution(Moves,Area) :-
	write('Solution '), write(Area), write(' {'),
	trace_moves(Moves),
	write('}'), nl,
	flush_output(user).
      
solution_quality([],0).
solution_quality([move(Size,_) | Moves],Area) :-
	solution_quality(Moves,Area1),
	Area is Area1 + Size * Size.

						% store solution
store_better_solution(Moves,Covered_area) :-
	solution(_,Old_area),
	Covered_area > Old_area,
	!,
	retract_old_solution,
	assert(solution(Moves,Covered_area)).

						% ignore solution
store_better_solution(_,_).

best_possible_solution(Covered_area,[],Covered_area).
best_possible_solution(Covered_area,[square(Size,_) | States],Max_area) :-
	Covered_area1 is Covered_area + Size * Size,
	best_possible_solution(Covered_area1,States,Max_area).
   
assess_solution(Covered_area,Moves) :-
	out_solution(Moves,Covered_area),
	store_better_solution(Moves,Covered_area).

trace_moves([]).
trace_moves([move(Size,p(X,Y)) | Moves]) :-
	trace_moves(Moves),
	write('{'),write(Size),write(' '),
	write(X),write(' '),write(Y),write('} ').

trace_moves1(Moves) :-
	trace_output(4),
	trace_moves(Moves),
	!.
trace_moves1(_).
   
trace_select_move(move(Size,p(X,Y))) :-
	trace_output(3),
	write(' '),write(Size),
	write('['),write(X),write(':'),write(Y),write(']'),
	!.
trace_select_move(_).
trace_state(State,Newstate) :-
	trace_output(1),
	write('old state     : '),write(State),nl,
	write('new state     : '),write(Newstate),nl,
	!.
trace_state(_,_).

check_better_move(Best,Max_possible) :-
	Best < Max_possible
	/* ,
	write('stil better solution possible '),nl,
	write('   currently best : '),write(Best),nl,
	write('   max possible   : '),write(Max_possible),nl */ .

/*
check_better_move(Best,Max_possible) :-
	Best >= Max_possible, 
	write('no better solution possible, trying another way'),nl,
	write('   currently best : '),write(Best),nl,
	write('   max possible   : '),write(Max_possible),nl,
	fail.
*/   

search([],Covered_area,_,Moves) :-
	assess_solution(Covered_area,Moves).

search(State,Covered_area,Square_size,Moves) :-
	select_move(State,Square_size,Move),
	trace_select_move(Move),
	satisfy_all_constraints(Move,State,Newstate),
	trace_state(State,Newstate),
	Move = move(Size,_),
	New_covered_area is Covered_area + Size * Size,
	best_possible_solution(New_covered_area,Newstate,Max_area),
	solution(_,Best_area),
	check_better_move(Best_area,Max_area),
	Moves1 = [Move | Moves],
	trace_moves1(Moves1),
	search(Newstate,New_covered_area,Size,Moves1).
	      

:-assert(solution([],0)).

						% write program version
test(_) :-
	write('Version {Square Puzzle Prolog Search Program $Revision: 1.1 $ $Date: 1996/07/11 13:21:15 $}'),nl,
	flush_output(user),
	fail.

						% init search
test(X) :-
	write('SearchSolutionsSquare '),write(X),nl,
	flush_output(user),
	retract_old_solution,
	assert(solution([],0)),
	initstate(X,State),
	search(State,0,X,[]),
	fail.

						% search finished
test(X) :-
	solution(_Squares, Area),
	write('BestSolution '),
	write(X), write(' '), write(Area), nl,
	flush_output(user).


						% test over an interval
						% of squares
test(X,Y) :-
	X > Y, !,fail.

test(X,Y) :-
	test(X),
	X1 is X + 1,
	test(X1,Y).
test(_,_) :- !.  

						% simulation of retractall

retract_old_solution :-
	retract(solution(_,_)),
	fail.

retract_old_solution.
