#!/bin/sh
# the next line restarts using wish (must be wish4.0 or newer) \
exec wish "$0" "$@"

# square covering visualisation
# with Tcl/Tk
# this program is a front end for the prolog search program
# square.pl
#
# version $Id: square.tcl,v 1.14 1996/07/11 18:24:07 uwe Exp $


wm title      . "Square Puzzle"
wm iconname   . "Square"
wm iconbitmap . "@./square.60x45.xbm"


# use full height of window
set height [expr [winfo screenheight .] - 30]
set width  [expr round($height * 1.2) ]
wm geometry   . "${width}x${height}+0+0"
wm protocol   . WM_DELETE_WINDOW quitSquare

unset height
unset width

# program menue -------------------------------------------------------------------

proc initMenu { } {
    frame .menu \
	-relief raised \
	-bd 1
    pack .menu -side top -fill x

    button .menu.logo \
	-relief flat \
	-text "Square Puzzle" \
	-fg IndianRed4 \
	-activeforeground IndianRed3 \
	-command "aboutSquarePuzzle"

    pack .menu.logo -side right

    set m  .menu.program
    set mm $m.m

    menubutton $m \
	-text "Program" \
	-menu $mm \
	-underline 0

    menu $mm
    $mm add command \
	-label "Reset" \
	-command initAll

    $mm add command \
	-label "Show" \
	-command lookupBestInHistory

    $mm add command \
	-label "Search" \
	-command openSquare

    $mm add command \
	-label "Non Stop Search" \
	-command nonStopSearch

    $mm add command \
	-label "Abort" \
	-command abortSquare \
	-state disabled

    $mm add separator
    $mm add command \
	-label "Hitlist" \
	-command "writeHitList"

    $mm add separator
    $mm add command \
	-label "Quit" \
	-command quitSquare

    set m .menu.search
    set mm $m.m

    menubutton $m \
	-text "Search" \
	-menu $mm \
	-underline 0

    menu $mm

    label .menu.sizeLab \
	-relief flat \
	-text "size:   " \
	-padx 3 \
	-fg IndianRed4

    label .menu.sizeMin \
	-relief flat \
	-text "   2" \
	-fg IndianRed4

    label .menu.sizeMax \
	-relief flat \
	-text "100" \
	-fg IndianRed4

    scale .menu.size \
	-from 2 \
	-to   100 \
	-width 12 \
	-length 200 \
	-bigincrement 5 \
	-showvalue false \
	-resolution 1 \
	-orient horizontal

    pack .menu.program .menu.search .menu.sizeLab .menu.sizeMin .menu.size .menu.sizeMax -side left
}

proc enableMenu { } {
    .menu.program.m entryconfigure 1 -state normal
    .menu.program.m entryconfigure 2 -state normal
    .menu.program.m entryconfigure 3 -state normal
    .menu.program.m entryconfigure 4 -state normal
    .menu.program.m entryconfigure 5 -state disabled

    .menu.size configure -state normal
}

proc disableMenu { } {
    .menu.program.m entryconfigure 1 -state disabled
    .menu.program.m entryconfigure 2 -state disabled
    .menu.program.m entryconfigure 3 -state disabled
    .menu.program.m entryconfigure 4 -state disabled
    .menu.program.m entryconfigure 5 -state normal

    .menu.size configure -state disabled
}

proc disableCompleteMenu { } {
    disableMenu
    .menu.program.m entryconfigure 5 -state disabled
}

proc setSize args {
    .menu.sizeLab configure -text "size:[format "%3d" $args]"
}

proc aboutSquarePuzzle { } {
    regsub -all {\$} { Square Puzzle

An Optimization Problem:

Cover as much as possible in a square of size N with all squares of \
size 1 to N-1

Author: Uwe Schmidt, uwe@fh-wedel.de
$Revision: 1.14 $
$Date: 1996/07/11 18:24:07 $
    } "" text1

    regsub {..:..:..} $text1 "" text1

    tk_dialog .about "About Square Puzzle" $text1 "@./square.120x120.xbm" 0 ok
}

#--------------------------------------------------------------------------------

proc initStatusLine { } {
    global plane

    frame .statusline \
	-relief raised \
	-borderwidth 1

    createLabel      .statusline.sizeLab       "square size"
    createCountLabel .statusline.sizeEntry     plane(len) 3
    createLabel      .statusline.areaLab       "area covered"
    createCountLabel .statusline.areaEntry     plane(areaCovered) 6
    createLabel      .statusline.of            "of"
    createCountLabel .statusline.size2Entry    plane(len2) 6
    createLabel      .statusline.percentLab    "in %"
    createCountLabel .statusline.percentEntry  plane(percentCovered) 5
    createLabel      .statusline.progLab       "curr search prog"
    createCountLabel .statusline.prog          plane(prog) 10

    label .statusline.status \
	-textvariable plane(status) \
	-padx 5 \
	-anchor e

    pack \
	.statusline.sizeLab    .statusline.sizeEntry \
	.statusline.areaLab    .statusline.areaEntry \
	.statusline.of         .statusline.size2Entry \
	.statusline.percentLab .statusline.percentEntry \
	.statusline.progLab    .statusline.prog \
	-side left \
	-fill x

    pack .statusline.status \
	-side left \
	-fill x \
	-expand yes

    pack .statusline -fill x -side bottom
}

proc createCountLabel {widgetName variable {width 5}} {
    label $widgetName
    $widgetName configure \
	    -textvariable $variable \
	    -width $width \
	    -anchor e \
	    -fg IndianRed4 \
	    -relief sunken \
	    -borderwidth 1
}

proc createLabel {widgetName text} {
    label $widgetName
    $widgetName configure \
	-text $text \
	-padx 5
}

proc setStatus mess {
    global plane

    catch {after cancel $plane(after)}
    set plane(after) ""

    set plane(status) $mess
    update idletasks
}

proc setDelayedStatus mess {
    global plane

    set plane(after) [after 2000 [list setStatus $mess]]
}

#--------------------------------------------------------------------------------

proc getSearchProgs {} {
    set progs ""
    foreach file [glob -nocomplain {square-*}] {
	if {[regexp {^square-[0-9A-Za-z]+$} $file] && [file executable $file]} {
	    lappend progs $file
	}
    }
    return [lsort $progs]
}

proc setSearchProg prog {
    global plane

    set plane(prog) $prog

    if {"$prog"==""} {
	disableCompleteMenu
	setStatus "can't find a search program \"square-*\""
	return
    }

    setStatus "search program set to \"$prog\""
    return
}

#--------------------------------------------------------------------------------

proc initProg { } {
    set progs [getSearchProgs]
    setSearchProg [lindex $progs 0]

    foreach p $progs {
	.menu.search.m add command \
	    -label $p \
	    -command "setSearchProg $p"
    }

    return
}

#--------------------------------------------------------------------------------

proc initPlane { } {
    global plane

    catch {destroy $plane(frame)}
    
    set plane(squareColors) { red1 green1 blue1 yellow1
	red2 green2 blue2 yellow2 
	red3 green3 blue3 yellow3 
    }
    set plane(squareColors)      indianred3

    set plane(frameColor)        darkblue
    set plane(squareForeground)  white
    set plane(squareBorderWidth) 2
    set plane(status)            ""
    set plane(squaresDrawn)      false
    set plane(nonStopSearch)     false

    set plane(frame) [frame .squareframe \
			  -borderwidth 0 \
			  -bg $plane(frameColor) \
			  -width 400 \
			  -height 400 \
			 ]

    .menu.size configure -variable plane(len)

    set plane(len)               2
    pack $plane(frame) -expand yes -fill both
}

proc squareColor { } {
    global plane

    # select first color from list and rotate list 
    set colors $plane(squareColors)
    set color  [lindex $colors 0]
    set plane(squareColors) [lrange $colors 1 end]
    lappend plane(squareColors) $color

    return $color
}

proc insertSquare {frame frameLen squareX squareY squareLen {squareColor red} {squareBorder 0}} {
    set square [label $frame.square_${squareX}_${squareY} \
		    -bg $squareColor \
		    -fg white \
		    -text $squareLen \
		    -borderwidth $squareBorder \
		    -relief raise]
    set x [expr double($squareX)   / double($frameLen)]
    set y [expr double($squareY)   / double($frameLen)]
    set l [expr double($squareLen) / double($frameLen)]

    place $square -in $frame -relx $x -rely $y -relwidth $l -relheight $l
}

proc drawSquare {squareLen squareX squareY} {
    global plane

    set frame    $plane(frame)
    set frameLen $plane(len)

    set square [label $frame.square_${squareX}_${squareY} \
		    -text        $squareLen \
		    -bg          [squareColor] \
		    -fg          $plane(squareForeground) \
		    -borderwidth $plane(squareBorderWidth) \
		    -relief      raise]

    set x [expr double($squareX)   / double($frameLen)]
    set y [expr double($squareY)   / double($frameLen)]
    set l [expr double($squareLen) / double($frameLen)]

    place $square -in $frame -relx $x -rely $y -relwidth $l -relheight $l
}

proc clearSquares { } {
    global plane

    if {$plane(squaresDrawn)} {
	eval destroy [winfo children $plane(frame)]
	set plane(squaresDrawn) false
    }
}

proc drawSolution { } {
    global plane

    clearSquares
    foreach s $plane(squareList) {
	eval drawSquare $s 
    }
    set plane(squaresDrawn) true
    update idletasks
}

#--------------------------------------------------------------------------------
#
# commands recieved from search program

proc Version vers {
    regsub {Revision: }  $vers "" vers
    regsub {Date: }      $vers "" vers
    regsub {..:..:..}    $vers "" vers
    regsub -all {\$}     $vers "" vers
    # wm title . $vers
}

proc SearchSolutionsSquare squareLen {
    global plane

    set plane(len)         $squareLen
    set plane(len2)        [expr $squareLen * $squareLen]

    clearSquares
}

proc Solution {covered squareList} {
    global plane

    # if new solution is found imediatly, cancel after cmd
    catch {after cancel $plane(after)}

    set oldArea $plane(areaCovered)

    set plane(areaCovered)          $covered
    set plane(squareList)           $squareList

    # store the whole solution
    insertHistory $plane(len) $covered $squareList $plane(prog)

    if {$oldArea == 0} {
	setStatus "1. solution"
    } else {
	setStatus "a better solution"
    }
    setDelayedStatus "searching..."

    drawSolution
}

proc BestSolution {len area} {
    global plane

    setStatus "best solution found"
    bell
    closeSquare
}

#--------------------------------------------------------------------------------
#
# history functions

proc insertHistory {len covered squares prog} {
    global history

    set history($len,$covered,$prog) $squares
    if {![info exists history($len,best,$prog)] || $history($len,best,$prog) < $covered} {
	set history($len,best,$prog) $covered
	storeHistory
    }

    if {![info exists history($len,best)] || $history($len,best) < $covered} {
	set history($len,$covered) $squares
	set history($len,best) $covered
	set history($len,bestprog) $prog
	storeHistory
    }

    if {![info exists history(max)] || $history(max) < $len} {
	set history(max) $len
	storeHistory
    }
}

proc lookupLargestInHistory { } {
    global history

    if {[info exists history(max)]} {
	return $history(max)
    }
    return 2
}

proc lookupBestInHistory {{len ""}} {
    global history plane

    if {"$len"!=""} {
	set plane(len) $len
    } else {
	set len $plane(len)
    }

    if {[info exists history($len,best)]} {
	set best $history($len,best)
	set mess "best known solution"
	if {[info exists history($len,bestprog)]} {
	    append mess ", search done by \"$history($len,bestprog)\""
	}
	set plane(areaCovered) $best
	set plane(squareList)  $history($len,$best)
	drawSolution
	setStatus $mess
    } else {
	setStatus "solution not yet computed"
    }

    update idletasks
    return
}

proc writeHitList { } {
    global history

    set max [lookupLargestInHistory]
    for {set i 2} {$i < $max} {incr i} {
	if {[info exists history($i,best)]} {
	    set covered $history($i,best)
	    set size2 [expr $i * $i]
	    set percent [expr double($covered) * 100.0 / double($size2)]
	    puts [format "size: %3d  covered: %8d  from %8d  in %% %3.2f" \
		      $i $covered $size2 $percent] \
	}
    }

}

proc storeHistory { } {
    global history

    set f [open "square.history" w]

    foreach i [lsort [array names history]] {
	puts $f [list set history($i) $history($i)]
    }

    close $f
}

proc loadHistory { } {
    global history

    if {[file readable "square.history"]} {
	source "square.history"
    }
}

#--------------------------------------------------------------------------------
#
# execute a run
# not used from menu

proc execSquare {} {
    global plane

    set prog $plane(prog)

    if {"$prog"==""} {
	setStatus "no search program defined"
	return
    }

    setStatus "searching..."
    update idletasks

    puts stderr [list exec nice $prog $plane(len)]
    set result [exec nice $prog $plane(len)]

    setStatus ""

    eval $result
}

#--------------------------------------------------------------------------------
#
# executes a run in background through a pipe and with event handlers

proc openSquare {} {
    global plane

    set prog $plane(prog)
    if {"$prog"==""} {
	setStatus "no search program defined"
	return
    }

    set plane(areaCovered) 0
    clearSquares
    setStatus "searching..."
    disableMenu
    update idletasks

    set plane(pipename) "/tmp/square-[pid]"

    # setStatus "openSquare: $plane(pipename)"

    catch {exec rm -f $plane(pipename)}

    if {[catch {exec mknod $plane(pipename) p}]} {
	setStatus "can't create pipe for communication with search program"
	return
    }

    # setStatus "pipe created"

    if {[catch {set plane(pipe) [open $plane(pipename) r+]}]} {
	setStatus "can't open pipe for communication with search program"
	return
    }

    # setStatus "pipe opened"

    puts stderr "start background process"
    puts stderr [list exec $prog $plane(len) > $plane(pipename) & ]

    if {[catch {set plane(processid) \
		    [ exec $prog $plane(len) > $plane(pipename) & ]}]} {
	setStatus "can't start search program \"$prog\""
	return
    }

    # setStatus "search program started: $plane(processid)"

    fileevent $plane(pipe) readable execSquareCmd
}

proc execSquareCmd {} {
    global plane

    if {[eof $plane(pipe)]} {
	closeSquare
    } else {
	gets $plane(pipe) line
	eval $line
	update idletasks
    }
}

proc abortSquare {} {
    global plane

    set plane(nonStopSearch) false
    setStatus "aborting search"

    closeSquare

    setStatus "search aborted"
}

proc closeSquare {} {
    global plane

    catch {close $plane(pipe)}
    catch {exec rm -f $plane(pipename)}
    catch {exec kill $plane(processid)}

    set plane(pipe) ""
    set plane(pipename) ""
    set plane(processid) ""

    if {$plane(nonStopSearch)} {
	# show final solution for 5 sec
	after 5000

	#start next size
	incr plane(len)
	openSquare
    } else {
	enableMenu
    }
}

proc nonStopSearch {} {
    global plane

    set plane(nonStopSearch) true
    openSquare
}

#--------------------------------------------------------------------------------
#
# update variables depending on plane(len) and plane(areaCovered)

trace variable plane(len) w updateLen

proc updateLen args {
    global plane

    .menu.sizeLab configure -text "size:[format "%3d" $plane(len)]"

    set plane(len2) [expr $plane(len) * $plane(len)]
    set plane(areaCovered) 0
    clearSquares
}

trace variable plane(areaCovered) w updateArea

proc updateArea args {
    global plane

    if {"$plane(areaCovered)"==""} {
	set plane(percentCovered) ""
    } else {
	set percent [expr double($plane(areaCovered)) * 100.0 / double($plane(len2))]
	set plane(percentCovered) [format "%2.2f" $percent]
    }
}

#--------------------------------------------------------------------------------
#
# program termination

proc quitSquare { } {

    # abort search, if running
    closeSquare

    catch {puts stdout "thanks for playing the square puzzle"}
    exit
}

#--------------------------------------------------------------------------------
#
# program initialization

proc initAll { } {
    global plane

    # just for reload
    catch {eval destroy [winfo children .]}

    # load old results
    loadHistory

    # initialize menu
    initMenu

    # initialize square frame
    initPlane

    # initialize status line
    initStatusLine

    # prepare Prolog search program
    initProg

    set plane(len) [lookupLargestInHistory]

    return
}

initAll